@@ -1,6 +1,7 @@
 # Author: Dimo Angelov
 #
 # License: BSD 3 clause
+import logging
 import numpy as np
 import pandas as pd
 from gensim.models.doc2vec import Doc2Vec, TaggedDocument
@@ -8,13 +9,23 @@ from gensim.utils import simple_preprocess
 from gensim.parsing.preprocessing import strip_tags
 import umap
 import hdbscan
-from operator import itemgetter
 from sklearn.metrics.pairwise import cosine_similarity
 from wordcloud import WordCloud
 import matplotlib.pyplot as plt
 from joblib import dump, load
 from sklearn.cluster import dbscan
 
+logger = logging.getLogger('top2vec')
+logger.setLevel(logging.WARNING)
+sh = logging.StreamHandler()
+sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
+logger.addHandler(sh)
+
+
+def default_tokenizer(doc):
+    """Tokenize documents for training and remove too long/short words"""
+    return simple_preprocess(strip_tags(doc), deacc=True)
+
 
 class Top2Vec:
     """
@@ -25,10 +36,10 @@ class Top2Vec:
 
     Parameters
     ----------
-    documents: list of str
+    documents: List of str
         Input corpus, should be a list of strings.
 
-    speed: string (optional, default 'fast-learn')
+    speed: string (Optional, default 'fast-learn')
         This parameter will determine how fast the model takes to train. The
         fast-learn option is the fastest and will generate the lowest quality
         vectors. The learn option will learn better quality vectors but take
@@ -39,37 +50,38 @@ class Top2Vec:
             * learn
             * deep-learn
 
-    workers: int (optional)
+    document_ids: List of str, int (Optional)
+        A unique value per document that will be used for referring to documents
+        in search results. If ids are not given, the index of each document
+        in the original corpus will become the id.
+
+    keep_documents: bool (Optional, default True)
+        If set to False documents will only be used for training and not saved
+        as part of the model. This will reduce model size. When using search
+        functions only document ids will be returned, not the actual documents.
+
+    workers: int (Optional)
         The amount of worker threads to be used in training the model. Larger
         amount will lead to faster training.
+    
+    tokenizer: callable or None (default)
+        Override the default tokenization method. If None then gensim.utils.simple_preprocess
+        will be used.
+    
+    verbose: bool (Optional, default False)
+        Whether to print status data during training.
 
-    Methods
-    -------
     """
 
-    def __init__(self, documents, speed="fast-learn", workers=None):
-        """
-        Parameters
-        ----------
-        documents: list of str
-            Input corpus, should be a list of strings.
+    def __init__(self, documents, speed="fast-learn", document_ids=None, keep_documents=True, workers=None,
+                 tokenizer=None, verbose=False):
 
-        speed: string (optional, default 'fast-learn')
-            This parameter will determine how fast the model takes to train. The
-            fast-learn option is the fastest and will generate the lowest quality
-            vectors. The learn option will learn better quality vectors but take
-            a longer time to train. The deep-learn option will learn the best quality
-            vectors but will take significant time to train. The valid string speed
-            options are:
-                * fast-learn
-                * learn
-                * deep-learn
+        if verbose:
+            logger.setLevel(logging.DEBUG)
+        else:
+            logger.setLevel(logging.WARNING)
 
-        workers: int (optional)
-            The amount of worker threads to be used in training the model. Larger
-            amount will lead to faster training.
-        """
-        # validate inputs
+        # validate training inputs
         if speed == "fast-learn":
             hs = 0
             negative = 5
@@ -82,6 +94,10 @@ class Top2Vec:
             hs = 1
             negative = 0
             epochs = 400
+        elif speed == "test-learn":
+            hs = 0
+            negative = 5
+            epochs = 1
         else:
             raise ValueError("speed parameter needs to be one of: fast-learn, learn or deep-learn")
 
@@ -92,13 +108,47 @@ class Top2Vec:
         else:
             raise ValueError("workers needs to be an int")
 
-        self.documents = list(documents)
+        if tokenizer is not None:
+            self._tokenizer = tokenizer
+        else:
+            self._tokenizer = default_tokenizer
 
-        # preprocess documents for training - tokenize and remove too long/short words
-        train_corpus = [TaggedDocument(simple_preprocess(strip_tags(doc), deacc=True), [i])
+        # validate documents
+        if not all((isinstance(doc, str) or isinstance(doc, np.str_)) for doc in documents):
+            raise ValueError("Documents need to be a list of strings")
+        if keep_documents:
+            self.documents = np.array(documents, dtype="object")
+        else:
+            self.documents = None
+
+        # validate document ids
+        if document_ids is not None:
+
+            if len(documents) != len(document_ids):
+                raise ValueError("Document ids need to match number of documents")
+            elif len(document_ids) != len(set(document_ids)):
+                raise ValueError("Document ids need to be unique")
+
+            if all((isinstance(doc_id, str) or isinstance(doc_id, np.str_)) for doc_id in document_ids):
+                self.doc_id_type = np.str_
+            elif all((isinstance(doc_id, int) or isinstance(doc_id, np.int_)) for doc_id in document_ids):
+                self.doc_id_type = np.int_
+            else:
+                raise ValueError("Document ids need to be str or int")
+
+            self.document_ids = np.array(document_ids)
+            self.doc_id2index = dict(zip(document_ids, list(range(0, len(document_ids)))))
+        else:
+            self.document_ids = None
+            self.doc_id2index = None
+            self.doc_id_type = np.int_
+
+        logger.info('Preprocessing documents for training')
+        train_corpus = [TaggedDocument(self._tokenizer(doc), [i])
                         for i, doc in enumerate(documents)]
 
         # create documents and word embeddings with doc2vec
+        logger.info('Creating joint document/word embedding')
         if workers is None:
             self.model = Doc2Vec(documents=train_corpus,
                                  vector_size=300,
@@ -123,16 +173,19 @@ class Top2Vec:
                                  dbow_words=1)
 
         # create 5D embeddings of documents
+        logger.info('Creating lower dimension embedding of documents')
         umap_model = umap.UMAP(n_neighbors=15,
                                n_components=5,
                                metric='cosine').fit(self.model.docvecs.vectors_docs)
 
         # find dense areas of document vectors
+        logger.info('Finding dense areas of documents')
         cluster = hdbscan.HDBSCAN(min_cluster_size=15,
                                   metric='euclidean',
                                   cluster_selection_method='eom').fit(umap_model.embedding_)
 
         # calculate topic vectors from dense areas of documents
+        logger.info('Finding topics')
         self._create_topic_vectors(cluster.labels_)
 
         # deduplicate topics
@@ -194,6 +247,8 @@ class Top2Vec:
         self.topic_words = []
         self.topic_word_scores = []
         np.apply_along_axis(self._generate_topic_words_scores, axis=1, arr=self.topic_vectors)
+        self.topic_words = np.array(self.topic_words)
+        self.topic_word_scores = np.array(self.topic_word_scores)
 
     def _generate_topic_words_scores(self, topic_vector):
         sim_words = self.model.wv.most_similar(positive=[topic_vector], topn=50)
@@ -231,8 +286,8 @@ class Top2Vec:
 
     def _validate_num_docs(self, num_docs):
         self._less_than_zero(num_docs, "num_docs")
-        document_count = len(self.documents)
-        if num_docs > document_count:
+        document_count = self.model.docvecs.count
+        if num_docs > self.model.docvecs.count:
             raise ValueError(f"num_docs cannot exceed the number of documents: {document_count}")
 
     def _validate_num_topics(self, num_topics):
@@ -247,29 +302,59 @@ class Top2Vec:
         if topic_num > topic_count:
             raise ValueError(f"Invalid topic number: valid topics numbers are 0 to {topic_count}")
 
-    def _validate_topic_search(self, topic_num, doc_num):
-        if doc_num > self.topic_sizes[topic_num]:
+    def _validate_topic_search(self, topic_num, num_docs):
+        if num_docs > self.topic_sizes[topic_num]:
             raise ValueError(f"Invalid number of documents: topic {topic_num}"
                              f" only has {self.topic_sizes[topic_num]} documents")
 
-    def _validate_doc_num(self, doc_num):
-        self._less_than_zero(doc_num, "doc_num")
-        document_count = len(self.documents) - 1
-        if doc_num > document_count:
-            raise ValueError(f"Invalid document number: valid document numbers are 0 to {document_count}")
+    def _validate_doc_ids(self, doc_ids, doc_ids_neg):
+
+        if not isinstance(doc_ids, list):
+            raise ValueError("doc_ids must be a list of string or int")
+
+        if not isinstance(doc_ids_neg, list):
+            raise ValueError("doc_ids_neg must be a list of string or int")
+
+        doc_ids_all = doc_ids + doc_ids_neg
+        for doc_id in doc_ids_all:
+            if self.document_ids is not None:
+                if doc_id not in self.document_ids:
+                    raise ValueError(f"{doc_id} is not a valid document id")
+            elif doc_id < 0 or doc_id > self.model.docvecs.count - 1:
+                raise ValueError(f"{doc_id} is not a valid document id")
 
     def _validate_keywords(self, keywords, keywords_neg):
 
-        if not isinstance(keywords, list):
+        if not (isinstance(keywords, list) or isinstance(keywords, np.ndarray)):
             raise ValueError("keywords must be a list of strings")
 
-        if not isinstance(keywords_neg, list):
+        if not (isinstance(keywords_neg, list) or isinstance(keywords_neg, np.ndarray)):
             raise ValueError("keywords_neg must be a list of strings")
 
-        for word in keywords + keywords_neg:
+        keywords_lower = [keyword.lower() for keyword in keywords]
+        keywords_neg_lower = [keyword.lower() for keyword in keywords_neg]
+
+        for word in keywords_lower + keywords_neg_lower:
             if word not in self.model.wv.vocab:
                 raise ValueError(f"'{word}' has not been learned by the model so it cannot be searched")
 
+        return keywords_lower, keywords_neg_lower
+
+    def _get_document_ids(self, doc_index):
+        if self.document_ids is None:
+            return doc_index
+        else:
+            return self.document_ids[doc_index]
+
+    def _get_document_indexes(self, doc_ids):
+        if self.document_ids is None:
+            return doc_ids
+        else:
+            return [self.doc_id2index[doc_id] for doc_id in doc_ids]
+
+    def _get_word_vectors(self, keywords):
+        return [self.model[word] for word in keywords]
+
     def get_num_topics(self):
         """
         Get number of topics.
@@ -291,12 +376,12 @@ class Top2Vec:
 
         Returns
         -------
-        topic_sizes: list
+        topic_sizes: array of int, shape(num_topics)
             The number of documents most similar to the topic.
-        topic_nums:
-            The unique index of every topic will be returned.
+        topic_nums: array of int, shape(num_topics)
+            The unique number of every topic will be returned.
         """
-        return self.topic_sizes.values.tolist(), self.topic_sizes.index.tolist()
+        return np.array(self.topic_sizes.values), np.array(self.topic_sizes.index)
 
     def get_topics(self, num_topics):
         """
@@ -309,12 +394,12 @@ class Top2Vec:
 
         Parameters
         ----------
-        num_topics: int
+        num_topics: int, shape(num_topics)
             Number of topics to return.
 
         Returns
         -------
-        topics_words: list, shape (n_topics, 50)
+        topics_words: array of shape(num_topics, 50)
             For each topic the top 50 words are returned, in order
             of semantic similarity to topic.
             Example:
@@ -322,7 +407,7 @@ class Top2Vec:
                  ['environment', 'warming', 'climate ... 'temperature']  <Topic 1>
                  ...]
 
-        word_scores: list, shape (n_topics, 50)
+        word_scores: array of shape(num_topics, 50)
             For each topic the cosine similarity scores of the
             top 50 words to the topic are returned.
             Example:
@@ -330,15 +415,15 @@ class Top2Vec:
                  [0.7818', 0.7671, 0.7603 ... 0.6769]  <Topic 1>
                  ...]
 
-        topic_nums: list of int
-            The unique index of every topic will be returned.
+        topic_nums: array of int, shape(num_topics)
+            The unique number of every topic will be returned.
         """
 
         self._validate_num_topics(num_topics)
 
-        return self.topic_words[0:num_topics], self.topic_word_scores[0:num_topics], list(range(0, num_topics))
+        return self.topic_words[0:num_topics], self.topic_word_scores[0:num_topics], np.array(range(0, num_topics))
 
-    def search_documents_by_topic(self, topic_num, num_docs):
+    def search_documents_by_topic(self, topic_num, num_docs, return_documents=True):
         """
         Get the most semantically similar documents to the topic.
 
@@ -354,33 +439,43 @@ class Top2Vec:
         num_docs: int
             Number of documents to return.
 
+        return_documents: bool (Optional default True)
+            Determines if the documents will be returned. If they were not saved
+            in the model they will also not be returned.
+
         Returns
         -------
-        documents: list of str
+        documents: (Optional) array of str, shape(num_docs)
             The documents in a list, the most similar are first.
 
-        doc_scores: float
+            Will only be returned if the documents were saved and if return_documents
+            is set to True.
+
+        doc_scores: array of float, shape(num_docs)
             Semantic similarity of document to topic. The cosine similarity of the
             document and topic vector.
 
-        doc_nums: list of int
-            Indexes of documents in the input corpus of documents.
+        doc_ids: array of int, shape(num_docs)
+            Unique ids of documents. If ids were not given, the index of document
+            in the original corpus will be returned.
         """
-
         self._validate_num_docs(num_docs)
         self._validate_topic_num(topic_num)
         self._validate_topic_search(topic_num, num_docs)
 
         topic_document_indexes = np.where(self.doc_top == topic_num)[0]
         topic_document_indexes_ordered = np.flip(np.argsort(self.doc_dist[topic_document_indexes]))
-        doc_nums = topic_document_indexes[topic_document_indexes_ordered][0:num_docs]
-        doc_scores = self.doc_dist[doc_nums]
+        doc_indexes = topic_document_indexes[topic_document_indexes_ordered][0:num_docs]
+        doc_scores = self.doc_dist[doc_indexes]
+        doc_ids = self._get_document_ids(doc_indexes)
 
-        documents = list(itemgetter(*doc_nums)(self.documents))
+        if self.documents is not None and return_documents:
+            documents = self.documents[doc_indexes]
+            return documents, doc_scores, doc_ids
+        else:
+            return doc_scores, doc_ids
 
-        return documents, doc_scores, doc_nums
-
-    def search_documents_by_keyword(self, keywords, num_docs, keywords_neg=[]):
+    def search_documents_by_keywords(self, keywords, num_docs, keywords_neg=[], return_documents=True):
         """
         Semantic search of documents using keywords.
 
@@ -393,42 +488,54 @@ class Top2Vec:
 
         Parameters
         ----------
-        keywords: list of str
+        keywords: List of str
             List of positive keywords being used for search of semantically similar
             documents.
 
-        keywords_neg: list of str (Optional)
+        keywords_neg: List of str (Optional)
             List of negative keywords being used for search of semantically dissimilar
             documents.
 
         num_docs: int
             Number of documents to return.
 
+        return_documents: bool (Optional default True)
+            Determines if the documents will be returned. If they were not saved
+            in the model they will also not be returned.
+
         Returns
         -------
-        documents: list of str
+        documents: (Optional) array of str, shape(num_docs)
             The documents in a list, the most similar are first.
 
-        doc_scores: list of float
+            Will only be returned if the documents were saved and if return_documents
+            is set to True.
+
+        doc_scores: array of float, shape(num_docs)
             Semantic similarity of document to keywords. The cosine similarity of the
             document and average of keyword vectors.
 
-        doc_nums: list of int
-            Indexes of documents in the input corpus of documents.
+        doc_ids: array of int, shape(num_docs)
+            Unique ids of documents. If ids were not given, the index of document
+            in the original corpus will be returned.
         """
         self._validate_num_docs(num_docs)
-        self._validate_keywords(keywords, keywords_neg)
+        keywords, keywords_neg = self._validate_keywords(keywords, keywords_neg)
 
-        word_vecs = [self.model[word] for word in keywords]
-        neg_word_vecs = [self.model[word] for word in keywords_neg]
+        word_vecs = self._get_word_vectors(keywords)
+        neg_word_vecs = self._get_word_vectors(keywords_neg)
         sim_docs = self.model.docvecs.most_similar(positive=word_vecs,
                                                    negative=neg_word_vecs,
                                                    topn=num_docs)
-        doc_nums = [doc[0] for doc in sim_docs]
-        doc_scores = [round(doc[1], 4) for doc in sim_docs]
-        documents = list(itemgetter(*doc_nums)(self.documents))
+        doc_indexes = [doc[0] for doc in sim_docs]
+        doc_scores = np.array([round(doc[1], 4) for doc in sim_docs])
+        doc_ids = self._get_document_ids(doc_indexes)
 
-        return documents, doc_scores, doc_nums
+        if self.documents is not None and return_documents:
+            documents = self.documents[doc_indexes]
+            return documents, doc_scores, doc_ids
+        else:
+            return doc_scores, doc_ids
 
     def similar_words(self, keywords, num_words, keywords_neg=[]):
         """
@@ -443,11 +550,11 @@ class Top2Vec:
 
         Parameters
         ----------
-        keywords: list of str
+        keywords: List of str
             List of positive keywords being used for search of semantically similar
             words.
 
-        keywords_neg: list of str
+        keywords_neg: List of str
             List of negative keywords being used for search of semantically dissimilar
             words.
 
@@ -457,22 +564,19 @@ class Top2Vec:
 
         Returns
         -------
-        words: list of str
+        words: array of str, shape(num_words)
             The words in a list, the most similar are first.
 
-        word_scores: list of float
+        word_scores: array of float, shape(num_words)
             Semantic similarity of word to keywords. The cosine similarity of the
             word and average of keyword vectors.
         """
-        self._validate_keywords(keywords, keywords_neg)
-
-        word_vecs = [self.model[word] for word in keywords]
-        neg_word_vecs = [self.model[word] for word in keywords_neg]
-        sim_words = self.model.wv.most_similar(positive=word_vecs,
-                                               negative=neg_word_vecs,
+        keywords, keywords_neg = self._validate_keywords(keywords, keywords_neg)
+        sim_words = self.model.wv.most_similar(positive=keywords,
+                                               negative=keywords_neg,
                                                topn=num_words)
-        words = [word[0] for word in sim_words]
-        word_scores = [round(word[1], 4) for word in sim_words]
+        words = np.array([word[0] for word in sim_words])
+        word_scores = np.array([round(word[1], 4) for word in sim_words])
 
         return words, word_scores
 
@@ -490,11 +594,11 @@ class Top2Vec:
 
         Parameters
         ----------
-        keywords: list of str
+        keywords: List of str
             List of positive keywords being used for search of semantically similar
             documents.
 
-        keywords_neg: (Optional) list of str
+        keywords_neg: (Optional) List of str
             List of negative keywords being used for search of semantically dissimilar
             documents.
 
@@ -503,31 +607,34 @@ class Top2Vec:
 
         Returns
         -------
-        topics_words: list, shape (n_topics, 50)
-            For each topic the top 50 words are returned, in order of semantic similarity to topic.
+        topics_words: array of shape (num_topics, 50)
+            For each topic the top 50 words are returned, in order of semantic
+            similarity to topic.
             Example:
                 [['data', 'deep', 'learning' ... 'artificial'],             <Topic 0>
                  ['environment', 'warming', 'climate ... 'temperature']     <Topic 1>
                  ...]
 
-        word_scores: list, shape (n_topics, 50)
-            For each topic the cosine similarity scores of the top 50 words to the topic are returned.
+        word_scores: array of shape (num_topics, 50)
+            For each topic the cosine similarity scores of the top 50 words
+            to the topic are returned.
             Example:
                 [[0.7132, 0.6473, 0.5700 ... 0.3455],     <Topic 0>
                  [0.7818', 0.7671, 0.7603 ... 0.6769]     <Topic 1>
                  ...]
 
-        topic_scores: list of float
-            For each topic the cosine similarity to the search keywords will be returned.
+        topic_scores: array of float, shape(num_topics)
+            For each topic the cosine similarity to the search keywords will be
+            returned.
 
-        topic_nums: list of int
-            The unique index of every topic will be returned.
+        topic_nums: array of int, shape(num_topics)
+            The unique number of every topic will be returned.
         """
         self._validate_num_topics(num_topics)
-        self._validate_keywords(keywords, keywords_neg)
+        keywords, keywords_neg = self._validate_keywords(keywords, keywords_neg)
 
-        word_vecs = [self.model[word] for word in keywords]
-        neg_word_vecs = [self.model[word] for word in keywords_neg]
+        word_vecs = self._get_word_vectors(keywords)
+        neg_word_vecs = self._get_word_vectors(keywords_neg)
 
         combined_vector = np.zeros(300, dtype=np.float64)
 
@@ -540,53 +647,74 @@ class Top2Vec:
         combined_vector /= (len(word_vecs) + len(neg_word_vecs))
 
         topic_ranks = [topic[0] for topic in cosine_similarity(self.topic_vectors, combined_vector.reshape(1, -1))]
-        topic_nums = list(np.flip(np.argsort(topic_ranks)[-num_topics:]))
+        topic_nums = np.flip(np.argsort(topic_ranks)[-num_topics:])
 
         topic_words = [self.topic_words[topic] for topic in topic_nums]
         word_scores = [self.topic_word_scores[topic] for topic in topic_nums]
-        topic_scores = [round(topic_ranks[topic], 4) for topic in topic_nums]
+        topic_scores = np.array([round(topic_ranks[topic], 4) for topic in topic_nums])
 
         return topic_words, word_scores, topic_scores, topic_nums
 
-    def search_documents_by_document(self, doc_num, num_docs):
+    def search_documents_by_documents(self, doc_ids, num_docs, doc_ids_neg=[], return_documents=True):
         """
-        Semantic similarity search of words.
+        Semantic similarity search of documents.
 
-        The most semantically similar documents to the document provided
-        will be returned. This method finds the closest document vectors
-        to the provided document vector.
+        The most semantically similar documents to the semantic combination of
+        document ids provided will be returned. If negative document ids are provided,
+        the documents will be semantically dissimilar to those document ids. Documents
+        will be ordered by decreasing similarity. This method finds the closest document
+        vectors to the provided documents averaged.
 
         Parameters
         ----------
-        doc_num: int
-            Index of document in the input corpus of documents.
+        doc_ids: List of int, str
+            Unique ids of document. If ids were not given, the index of document
+            in the original corpus.
+
+        doc_ids_neg: (Optional) List of int, str
+            Unique ids of document. If ids were not given, the index of document
+            in the original corpus.
 
         num_docs: int
             Number of documents to return.
 
+        return_documents: bool (Optional default True)
+            Determines if the documents will be returned. If they were not saved
+            in the model they will also not be returned.
+
         Returns
         -------
-        documents: list of str
+        documents: (Optional) array of str, shape(num_docs)
             The documents in a list, the most similar are first.
 
-        doc_scores: list of float
+            Will only be returned if the documents were saved and if return_documents
+            is set to True.
+
+        doc_scores: array of float, shape(num_docs)
             Semantic similarity of document to keywords. The cosine similarity of the
             document and average of keyword vectors.
 
-        doc_nums: list of int
-            Indexes of documents in the input corpus of documents.
+        doc_ids: array of int, shape(num_docs)
+            Unique ids of documents. If ids were not given, the index of document
+            in the original corpus will be returned.
         """
         self._validate_num_docs(num_docs)
-        self._validate_doc_num(doc_num)
+        self._validate_doc_ids(doc_ids, doc_ids_neg)
 
-        message_vec = self.model.docvecs[doc_num]
-        sim_docs = self.model.docvecs.most_similar(positive=[message_vec],
+        doc_indexes = self._get_document_indexes(doc_ids)
+        doc_indexes_neg = self._get_document_indexes(doc_ids_neg)
+        sim_docs = self.model.docvecs.most_similar(positive=doc_indexes,
+                                                   negative=doc_indexes_neg,
                                                    topn=num_docs)
-        doc_nums = [doc[0] for doc in sim_docs]
-        doc_scores = [round(doc[1], 4) for doc in sim_docs]
-        documents = list(itemgetter(*doc_nums)(self.documents))
+        doc_indexes = [doc[0] for doc in sim_docs]
+        doc_scores = np.array([round(doc[1], 4) for doc in sim_docs])
+        doc_ids = self._get_document_ids(doc_indexes)
 
-        return documents, doc_scores, doc_nums
+        if self.documents is not None and return_documents:
+            documents = self.documents[doc_indexes]
+            return documents, doc_scores, doc_ids
+        else:
+            return doc_scores, doc_ids
 
     def generate_topic_wordcloud(self, topic_num, background_color="black"):
         """
